/*
 * The Clear BSD License
 * Copyright 2017 NXP
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted (subject to the limitations in the disclaimer below) provided
 *  that the following conditions are met:
 *
 * o Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 *
 * o Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 *
 * o Neither the name of the copyright holder nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY THIS LICENSE.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "fsl_common.h"
#include "fsl_debug_console.h"
#include "board.h"
#include "fsl_gpio.h"

/*******************************************************************************
 * Code
 ******************************************************************************/
 
/* Get debug console frequency. */
uint32_t BOARD_DebugConsoleSrcFreq(void)
{
    uint32_t freq;

    /* To make it simple, we assume default PLL and divider settings, and the only variable
       from application is use PLL3 source or OSC source */
    if (CLOCK_GetMux(kCLOCK_UartMux) == 0) /* PLL3 div6 80M */
    {
        freq = (CLOCK_GetPllFreq(kCLOCK_PllUsb1) / 6U) / (CLOCK_GetDiv(kCLOCK_UartDiv) + 1U);
    }
    else
    {
        freq = CLOCK_GetOscFreq() / (CLOCK_GetDiv(kCLOCK_UartDiv) + 1U);
    }

    return freq;
} 
 
/* Initialize debug console. */
void BOARD_InitDebugConsole(void)
{
    uint32_t uartClkSrcFreq = BOARD_DebugConsoleSrcFreq();

    DbgConsole_Init(BOARD_DEBUG_UART_BASEADDR, BOARD_DEBUG_UART_BAUDRATE, BOARD_DEBUG_UART_TYPE, uartClkSrcFreq);
}
   
/* MPU configuration. */
void BOARD_ConfigMPU(void)
{
    /* Disable I cache and D cache */ 
    SCB_DisableICache();
    SCB_DisableDCache();
	    
    /* Disable MPU */ 
    ARM_MPU_Disable();

    /* Region 0 setting */    
    MPU->RBAR = ARM_MPU_RBAR(0, 0xC0000000U);
    MPU->RASR = ARM_MPU_RASR(0, ARM_MPU_AP_FULL, 2, 0, 0, 0, 0, ARM_MPU_REGION_SIZE_512MB);

    /* Region 1 setting */
    MPU->RBAR = ARM_MPU_RBAR(1, 0x80000000U);
    MPU->RASR = ARM_MPU_RASR(0, ARM_MPU_AP_FULL, 2, 0, 0, 0, 0, ARM_MPU_REGION_SIZE_1GB);

    /* Region 2 setting */
#if defined(XIP_EXTERNAL_FLASH)    
    MPU->RBAR = ARM_MPU_RBAR(2, 0x60000000U);
    MPU->RASR = ARM_MPU_RASR(0, ARM_MPU_AP_FULL, 0, 0, 1, 1, 0, ARM_MPU_REGION_SIZE_512MB);
#else
    MPU->RBAR = ARM_MPU_RBAR(2, 0x60000000U);
    MPU->RASR = ARM_MPU_RASR(0, ARM_MPU_AP_FULL, 2, 0, 0, 0, 0, ARM_MPU_REGION_SIZE_512MB);
#endif    

    /* Region 3 setting */
    MPU->RBAR = ARM_MPU_RBAR(3, 0x00000000U);
    MPU->RASR = ARM_MPU_RASR(0, ARM_MPU_AP_FULL, 2, 0, 0, 0, 0, ARM_MPU_REGION_SIZE_1GB);

    /* Region 4 setting */
    MPU->RBAR = ARM_MPU_RBAR(4, 0x00000000U);
    MPU->RASR = ARM_MPU_RASR(0, ARM_MPU_AP_FULL, 0, 0, 1, 1, 0, ARM_MPU_REGION_SIZE_64KB);

    /* Region 5 setting */
    MPU->RBAR = ARM_MPU_RBAR(5, 0x20000000U);
    MPU->RASR = ARM_MPU_RASR(0, ARM_MPU_AP_FULL, 0, 0, 1, 1, 0, ARM_MPU_REGION_SIZE_32KB);    

    /* Region 6 setting */
    MPU->RBAR = ARM_MPU_RBAR(6, 0x20200000U);
    MPU->RASR = ARM_MPU_RASR(0, ARM_MPU_AP_FULL, 1, 0, 0, 0, 0, ARM_MPU_REGION_SIZE_256KB);    

    /* Region 7 setting, set whole SDRAM can be accessed by cache */
    MPU->RBAR = ARM_MPU_RBAR(7, 0x80000000U);
    MPU->RASR = ARM_MPU_RASR(0, ARM_MPU_AP_FULL, 0, 0, 1, 1, 0, ARM_MPU_REGION_SIZE_32MB);    

    /* Region 8 setting, set last 2MB of SDRAM can't be accessed by cache, glocal variables which are not expected to be accessed by cache can be put here */
    MPU->RBAR = ARM_MPU_RBAR(8, 0x81E00000U);
    MPU->RASR = ARM_MPU_RASR(0, ARM_MPU_AP_FULL, 1, 0, 0, 0, 0, ARM_MPU_REGION_SIZE_2MB);   

    /* Enable MPU */ 
    ARM_MPU_Enable(MPU_CTRL_PRIVDEFENA_Msk);
	
    /* Enable I cache and D cache */ 
    SCB_EnableDCache();
    SCB_EnableICache();
}

void BOARD_SPILCD_RstPinInit(void)
{
    gpio_pin_config_t gpioPinConfig = {
        .direction = kGPIO_DigitalOutput,
        .outputLogic = 1u,
        .interruptMode = kGPIO_NoIntmode
    };
    GPIO_PinInit(BOARD_SPILCD_RST_GPIO_BASE, BOARD_SPILCD_RST_PIN_INDEX, &gpioPinConfig);
}

void BOARD_SPILCD_DcPinInit(void)
{
    gpio_pin_config_t gpioPinConfig = {
        .direction = kGPIO_DigitalOutput,
        .outputLogic = 1u,
        .interruptMode = kGPIO_NoIntmode
    };
    GPIO_PinInit(BOARD_SPILCD_DC_GPIO_BASE, BOARD_SPILCD_DC_PIN_INDEX, &gpioPinConfig);
}

void BOARD_SPILCD_CsPinInit(void)
{
    gpio_pin_config_t gpioPinConfig = {
        .direction = kGPIO_DigitalOutput,
        .outputLogic = 1u,
        .interruptMode = kGPIO_NoIntmode
    };
    GPIO_PinInit(BOARD_SPILCD_CS_GPIO_BASE, BOARD_SPILCD_CS_PIN_INDEX, &gpioPinConfig);
}

void BOARD_SPILCD_pullRstPin(bool value)
{
    if(value)
    {
        BOARD_SPILCD_RST_GPIO_BASE->DR |=  1<<BOARD_SPILCD_RST_PIN_INDEX;
    }
    else
    {
        BOARD_SPILCD_RST_GPIO_BASE->DR &= ~(1<<BOARD_SPILCD_RST_PIN_INDEX);
    }
}

void BOARD_SPILCD_pullDcPin(bool value)
{
    if(value)
    {
        BOARD_SPILCD_DC_GPIO_BASE->DR |=  1<<BOARD_SPILCD_DC_PIN_INDEX;
    }
    else
    {
        BOARD_SPILCD_DC_GPIO_BASE->DR &= ~(1<<BOARD_SPILCD_DC_PIN_INDEX);
    }
}

void BOARD_SPILCD_pullCsPin(bool value)
{
    if(value)
    {
        BOARD_SPILCD_CS_GPIO_BASE->DR |= 1<<BOARD_SPILCD_CS_PIN_INDEX;
    }
    else
    {
        BOARD_SPILCD_CS_GPIO_BASE->DR &= ~(1<<BOARD_SPILCD_CS_PIN_INDEX);
    }
}

void BOARD_Camera_RstPinInit(void)
{
    gpio_pin_config_t gpioPinConfig = {
        .direction = kGPIO_DigitalOutput,
        .outputLogic = 1u,
        .interruptMode = kGPIO_NoIntmode
    };
    GPIO_PinInit(BOARD_CAMERA_RST_GPIO_BASE, BOARD_CAMERA_RST_PIN_INDEX, &gpioPinConfig);
}

void BOARD_Camera_PdwnPinInit(void)
{
    gpio_pin_config_t gpioPinConfig = {
        .direction = kGPIO_DigitalOutput,
        .outputLogic = 0u,
        .interruptMode = kGPIO_NoIntmode
    };
    GPIO_PinInit(BOARD_CAMERA_PDWN_GPIO_BASE, BOARD_CAMERA_PDWN_PIN_INDEX, &gpioPinConfig);
}

void BOARD_Camera_VsynPinInit(void)
{
    gpio_pin_config_t gpioPinConfig = {
        .direction = kGPIO_DigitalInput,
        .interruptMode = kGPIO_IntRisingEdge
    };
    GPIO_PinInit(BOARD_CAMERA_VSYNC_GPIO_BASE, BOARD_CAMERA_VSYNC_PIN_INDEX, &gpioPinConfig);
    GPIO_EnableInterrupts(BOARD_CAMERA_VSYNC_GPIO_BASE, 1u << BOARD_CAMERA_VSYNC_PIN_INDEX);
    EnableIRQ(BOARD_CAMERA_VSYNC_IRQn);
}

void BOARD_Camera_pullRstPin(bool value)
{
    if(value)
    {
        GPIO_SetPinsOutput(BOARD_CAMERA_RST_GPIO_BASE, 1<<BOARD_CAMERA_RST_PIN_INDEX);
    }
    else
    {
        GPIO_ClearPinsOutput(BOARD_CAMERA_RST_GPIO_BASE, 1<<BOARD_CAMERA_RST_PIN_INDEX);
    }
}

void BOARD_Camera_pullPdwnPin(bool value)
{
    if(value)
    {
        GPIO_SetPinsOutput(BOARD_CAMERA_PDWN_GPIO_BASE, 1<<BOARD_CAMERA_PDWN_PIN_INDEX);
    }
    else
    {
        GPIO_ClearPinsOutput(BOARD_CAMERA_PDWN_GPIO_BASE, 1<<BOARD_CAMERA_PDWN_PIN_INDEX);
    }
}


